﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using Attributes;
  using Interpreter;
  using Persistence.Default.CompositeSerializers.Storable;
  using Stack;

  [PushExpression(StackTypes.Char, "CHAR.ISWHITESPACE", StackTypes.Boolean)]
  [StorableClass]
  public class CharIsWhitespaceExpression : StatelessExpression {
    public CharIsWhitespaceExpression() { }
    [StorableConstructor]
    public CharIsWhitespaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var c = interpreter.CharStack.Pop();
      interpreter.BooleanStack.Push(char.IsWhiteSpace(c));
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.ISLETTER", StackTypes.Boolean)]
  [StorableClass]
  public class CharIsLetterExpression : StatelessExpression {
    public CharIsLetterExpression() { }
    [StorableConstructor]
    public CharIsLetterExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var c = interpreter.CharStack.Pop();
      interpreter.BooleanStack.Push(char.IsLetter(c));
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.ISDIGIT", StackTypes.Boolean)]
  [StorableClass]
  public class CharIsDigitExpression : StatelessExpression {
    public CharIsDigitExpression() { }
    [StorableConstructor]
    public CharIsDigitExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var c = interpreter.CharStack.Pop();
      interpreter.BooleanStack.Push(char.IsDigit(c));
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.FROMINTEGER", StackTypes.Integer)]
  [StorableClass]
  public class CharFromIntegerExpression : StatelessExpression {
    public CharFromIntegerExpression() { }
    [StorableConstructor]
    public CharFromIntegerExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = (int)Math.Abs(interpreter.IntegerStack.Pop() % 128);
      var c = Convert.ToChar(value);

      interpreter.CharStack.Push(c);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.FROMFLOAT", StackTypes.Float)]
  [StorableClass]
  public class CharFromFloatExpression : StatelessExpression {
    public CharFromFloatExpression() { }
    [StorableConstructor]
    public CharFromFloatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = (int)Math.Abs(interpreter.FloatStack.Pop() % 128);
      var c = Convert.ToChar(value);

      interpreter.CharStack.Push(c);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.ALLFROMSTRING", StackTypes.String)]
  [StorableClass]
  public class CharAllFromStringExpression : StatelessExpression {
    public CharAllFromStringExpression() { }
    [StorableConstructor]
    public CharAllFromStringExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var chars = interpreter.StringStack.Pop().ToCharArray();
      interpreter.CharStack.Push(chars);
    }
  }

}
