﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using System.Collections.Generic;

  using Attributes;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

  using Interpreter;
  using Stack;

  /// <summary>
  ///     Defines the name on top of the NAME stack as an instruction that will push the top item of the T stack onto the
  ///     EXEC stack.
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  [StorableClass]
  public abstract class DefineExpression<T> : StatelessExpression {
    protected DefineExpression() { }
    [StorableConstructor]
    protected DefineExpression(bool deserializing) : base(deserializing) { }

    protected void Eval(
      IPushStack<T> stack,
      IPushStack<string> nameStack,
      IDictionary<string, Expression> customExpressions,
      Func<T, Expression> creator) {
      var name = nameStack.Pop();
      var expression = creator(stack.Top);

      if (customExpressions.ContainsKey(name)) customExpressions[name] = expression;
      else customExpressions.Add(name, expression);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class CodeDefineExpression : DefineExpression<Expression> {
    public CodeDefineExpression() { }
    [StorableConstructor]
    protected CodeDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return (interpreter.Configuration.TopLevelPushCode && interpreter.CodeStack.Count < 2) ||
             interpreter.CodeStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.CodeStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new ExecPushExpression(v));

      interpreter.CodeStack.Pop();
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class ExecDefineExpression : DefineExpression<Expression> {
    public ExecDefineExpression() { }
    [StorableConstructor]
    protected ExecDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.ExecStack.Count < 2 ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.ExecStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new ExecPushExpression(v));

      interpreter.ExecStack.Pop();
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class FloatDefineExpression : DefineExpression<double> {
    public FloatDefineExpression() { }
    [StorableConstructor]
    protected FloatDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.FloatStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new FloatPushExpression(v));
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class IntegerDefineExpression : DefineExpression<long> {
    public IntegerDefineExpression() { }
    [StorableConstructor]
    protected IntegerDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.IntegerStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new IntegerPushExpression(v));
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class BooleanDefineExpression : DefineExpression<bool> {
    public BooleanDefineExpression() { }
    [StorableConstructor]
    protected BooleanDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.BooleanStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new BooleanPushExpression(v));
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class CharDefineExpression : DefineExpression<char> {
    public CharDefineExpression() { }
    [StorableConstructor]
    protected CharDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.CharStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new CharPushExpression(v));
    }
  }

  [PushExpression(StackTypes.String, "STRING.DEFINE", StackTypes.Name)]
  [StorableClass]
  public class StringDefineExpression : DefineExpression<string> {
    public StringDefineExpression() { }
    [StorableConstructor]
    protected StringDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.StringStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new StringPushExpression(v));
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].DEFINE", StackTypes.Name)]
  [StorableClass]
  public class IntegerVectorDefineExpression : DefineExpression<List<long>> {
    public IntegerVectorDefineExpression() { }
    [StorableConstructor]
    protected IntegerVectorDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerVectorStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.IntegerVectorStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new IntegerVectorPushExpression(v));
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].DEFINE", StackTypes.Name)]
  [StorableClass]
  public class FloatVectorDefineExpression : DefineExpression<List<double>> {
    public FloatVectorDefineExpression() { }
    [StorableConstructor]
    protected FloatVectorDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatVectorStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.FloatVectorStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new FloatVectorPushExpression(v));
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].DEFINE", StackTypes.Name)]
  [StorableClass]
  public class BooleanVectorDefineExpression : DefineExpression<List<bool>> {
    public BooleanVectorDefineExpression() { }
    [StorableConstructor]
    protected BooleanVectorDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanVectorStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.BooleanVectorStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new BooleanVectorPushExpression(v));
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].DEFINE", StackTypes.Name)]
  [StorableClass]
  public class StringVectorDefineExpression : DefineExpression<List<string>> {
    public StringVectorDefineExpression() { }
    [StorableConstructor]
    protected StringVectorDefineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringVectorStack.IsEmpty ||
             interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(
        interpreter.StringVectorStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new StringVectorPushExpression(v));
    }
  }
}