﻿using System;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Linq;

  using Common;
  using Configuration;
  using Core;
  using Data.Pool;
  using HeuristicLab.Data;
  using HeuristicLab.Encodings.IntegerVectorEncoding;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Problem.BenchmarkSuite;

  using Interpreter;
  using Optimization;
  using Parameters;
  using Persistence.Default.CompositeSerializers.Storable;
  using Random;

  [StorableClass]
  public abstract class PushProblem : SingleObjectiveBasicProblem<IntegerVectorEncoding> {
    [Storable]
    protected readonly PushConfiguration config;
    protected PushInterpreterPool pool;
    protected readonly ObjectPool<IRandom> randomPool = new ObjectPool<IRandom>(() => new MersenneTwister());

    [Storable]
    protected readonly IPushEvaluator PushEvaluator;

    public const string CaseQualitiesScopeParameterName = "CaseQualities";
    private const string BestTrainingSolutionResultName = "Best Solution";
    private const string TestQualityResultName = "Test Quality";

    protected PushProblem(PushBenchmarkSuiteEvaluator evaluator) {
      config = new PushConfiguration();
      PushEvaluator = evaluator;

      InitData();
      InitEvents();
      InitParameters();
    }

    [StorableConstructor]
    protected PushProblem(bool deserializing)
      : base(deserializing) {
    }

    protected PushProblem(PushProblem original, Cloner cloner)
      : base(original, cloner) {
      config = cloner.Clone(original.config);
      PushEvaluator = cloner.Clone(original.PushEvaluator);

      InitData();
      InitEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitData();
      InitEvents();
    }

    private void InitData() {
      pool = new PushInterpreterPool(Environment.ProcessorCount * 2, 4096, 1024, config);
    }

    private void InitEvents() {
      config.EnabledExpressionsChanged += EnabledExpressionsChanged;
    }

    private void EnabledExpressionsChanged(object sender, EnabledExpressionsChangedEventArgs e) {
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.BoundsParameter.Value[0, 1] = config.EnabledExpressions.Count;
    }

    #region Parameters

    private const string InstructionsParameterName = "Instructions";
    private const string InstructionsParameterDescription = "Enables/Disables Instructions";
    private const string EvalPushLimitParameterName = "EvalPushLimit";
    private const string EvalPushLimitParameterDescription = "This is the maximum allowed number of \"executions\" in a single top-level call to the interpreter. The execution of a single Push instruction counts as one execution, as does the processing of a single literal, as does the descent into one layer of parentheses (that is, the processing of the \"(\" counts as one execution).";
    private const string MinPointsInProgramParameterName = "MinProgramLength";
    private const string MinProgramLengthParameterDescription = "This is the minimum size of an item on the CODE/EXEC stack, expressed as a number of points. A point is an instruction, a literal, or a pair of parentheses.";
    private const string MaxPointsInProgramParameterName = "MaxProgramLength";
    private const string MaxProgramLengthParameterDescription = "This is the maximum size of an item on the CODE/EXEC stack, expressed as a number of points. A point is an instruction, a literal, or a pair of parentheses.";
    private const string TopLevelPushCodeParameterName = "TopLevelPushCode";
    private const string TopLevelPushCodeParameterDescription = "When TRUE (which is the default), code passed to the top level of the interpreter will be pushed onto the CODE stack prior to execution.";
    private const string TopLevelPopCodeParameterName = "TopLevelPopCode";
    private const string TopLevelPopCodeParameterDescription = "When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.";
    private const string MaxPointsInRandomInstructionParameterName = "MaxPointsInRandomInstruction";
    private const string MaxPointsInRandomInstructionParameterDescription = "MaxPointsInRandomInstruction";
    private const string ErcOptionsParameterName = "ERC options";
    private const string MaxStringLengthParameterName = "Max. string length";
    private const string MaxDepthParameterName = "Max. program recursion";

    private void InitParameters() {
      Parameters.Add(new ValueParameter<IEnabledExpressionsConfiguration>(
        InstructionsParameterName,
        InstructionsParameterDescription,
        config));

      Parameters.Add(new ValueParameter<ErcOptions>(ErcOptionsParameterName, config.ErcOptions));

      Parameters.Add(new FixedValueParameter<IntValue>(
        EvalPushLimitParameterName,
        EvalPushLimitParameterDescription,
        new IntValue(config.EvalPushLimit)));

      Parameters.Add(new FixedValueParameter<IntValue>(
        MinPointsInProgramParameterName,
        MinProgramLengthParameterDescription,
        new IntValue(config.MinPointsInProgram)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxPointsInProgramParameterName,
        MaxProgramLengthParameterDescription,
        new IntValue(config.MaxPointsInProgram)));
      Encoding.LengthParameter = MaxPointsInProgramParameter as IFixedValueParameter<IntValue>;

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPushCodeParameterName,
        TopLevelPushCodeParameterDescription,
        new BoolValue(config.TopLevelPushCode)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPopCodeParameterName,
        TopLevelPopCodeParameterDescription,
        new BoolValue(config.TopLevelPopCode)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxPointsInRandomInstructionParameterName,
        MaxPointsInRandomInstructionParameterDescription,
        new IntValue(config.MaxPointsInRandomExpression)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxStringLengthParameterName,
        new IntValue(config.MaxStringLength)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxDepthParameterName,
        new IntValue(config.MaxDepth)) { Hidden = true });

      Parameters.Add(new LookupParameter<BoolArray>("Cases", "The training cases that have been successfully executed."));
      Parameters.Add(new LookupParameter<DoubleArray>(CaseQualitiesScopeParameterName, "The quality of every single training case for each individual"));

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.Length = config.MaxPointsInProgram;
    }

    public IValueParameter<IEnabledExpressionsConfiguration> InstructionsParameter
    {
      get { return (IValueParameter<IEnabledExpressionsConfiguration>)Parameters[InstructionsParameterName]; }
    }

    public IEnabledExpressionsConfiguration Instructions
    {
      get { return InstructionsParameter.Value; }
      set { InstructionsParameter.Value = value; }
    }

    public IValueParameter<ErcOptions> ErcOptionsParameter
    {
      get { return (IValueParameter<ErcOptions>)Parameters[ErcOptionsParameterName]; }
    }

    public ErcOptions ErcOptions
    {
      get { return config.ErcOptions; }
      set
      {
        ErcOptionsParameter.Value = value;
        config.ErcOptions = value;
      }
    }

    /// <summary>
    ///     This is the maximum allowed number of "executions" in a single top-level call to the interpreter.
    ///     The execution of a single Push instruction counts as one execution, as does the processing of a single literal,
    ///     as does the descent into one layer of parentheses (that is, the processing of the "(" counts as one execution).
    ///     When this limit is exceeded the interpreter aborts immediately, leaving its stacks in the states they were in prior
    ///     to the abort (so they may still be examined by a calling program). Whether or not this counts as an "abnormal"
    ///     termination
    ///     is up to the calling program.
    /// </summary>
    public IValueParameter<IntValue> EvalPushLimitParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[EvalPushLimitParameterName]; }
    }

    public int EvalPushLimit
    {
      get { return config.EvalPushLimit; }
      set
      {
        EvalPushLimitParameter.Value.Value = value;
        config.EvalPushLimit = value;
      }
    }

    /// <summary>
    /// This is the maximum of depth a push program can have. Expressions, which lead to exceed this limit are interpreted as NOOP.
    /// </summary>
    public IValueParameter<IntValue> MaxDepthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxDepthParameterName]; }
    }

    public int MaxDepth
    {
      get { return config.MaxDepth; }
      set
      {
        MaxDepthParameter.Value.Value = value;
        config.MaxDepth = value;
      }
    }

    /// <summary>
    ///     This is the maximum size of an item on the CODE stack, expressed as a number of points.
    ///     A point is an instruction, a literal, or a pair of parentheses. Any instruction that would cause this limit to be
    ///     exceeded
    ///     should instead act as a NOOP, leaving all stacks in the states that they were in before the execution of the
    ///     instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInProgramParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInProgramParameterName]; }
    }

    public int MaxPointsInProgram
    {
      get { return config.MaxPointsInProgram; }
      set
      {
        MaxPointsInProgramParameter.Value.Value = value;
        config.MaxPointsInProgram = value;
      }
    }

    /// <summary>
    ///     This is the minimum size of an item on the CODE stack, expressed as a number of points.
    ///     A point is an instruction, a literal, or a pair of parentheses. Any instruction that would cause this limit to be
    ///     exceeded
    ///     should instead act as a NOOP, leaving all stacks in the states that they were in before the execution of the
    ///     instruction.
    /// </summary>
    public IValueParameter<IntValue> MinPointsInProgramParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MinPointsInProgramParameterName]; }
    }

    public int MinPointsInProgram
    {
      get { return config.MaxPointsInProgram; }
      set
      {
        MinPointsInProgramParameter.Value.Value = value;
        config.MinPointsInProgram = value;
      }
    }

    /// <summary>
    ///     The maximum number of points in an expression produced by the CODE.RAND instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInRandomExpressionParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInRandomInstructionParameterName]; }
    }

    public int MaxPointsInRandomExpression
    {
      get { return config.MaxPointsInRandomExpression; }
      set
      {
        MaxPointsInRandomExpressionParameter.Value.Value = value;
        config.MaxPointsInRandomExpression = value;
      }
    }

    /// <summary>
    ///     When TRUE (which is the default), code passed to the top level of the interpreter
    ///     will be pushed onto the CODE stack prior to execution.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPushCodeParameter
    {
      get { return (IValueParameter<BoolValue>)Parameters[TopLevelPushCodeParameterName]; }
    }

    public bool TopLevelPushCode
    {
      get { return config.TopLevelPushCode; }
      set
      {
        TopLevelPushCodeParameter.Value.Value = value;
        config.TopLevelPushCode = value;
      }
    }

    /// <summary>
    ///     When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPopCodeParameter
    {
      get { return (IValueParameter<BoolValue>)Parameters[TopLevelPopCodeParameterName]; }
    }

    public bool TopLevelPopCode
    {
      get { return config.TopLevelPopCode; }
      set
      {
        TopLevelPopCodeParameter.Value.Value = value;
        config.TopLevelPopCode = value;
      }
    }

    public IValueParameter<IntValue> MaxStringLengthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxStringLengthParameterName]; }
    }

    public int MaxStringLength
    {
      get { return config.MaxStringLength; }
      set
      {
        MaxStringLengthParameter.Value.Value = value;
        config.MaxStringLength = value;
      }
    }
    #endregion

    public override bool Maximization { get { return false; } }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      var bestQuality = Maximization ? qualities.Max() : qualities.Min();
      var bestIdx = Array.IndexOf(qualities, bestQuality);
      var bestIndividual = individuals[bestIdx].IntegerVector();

      var isIndividualBetter = AnalyzeBestTrainingSolution(bestIndividual, bestQuality, results, random);

      if (isIndividualBetter) {
        AnalyzeBestTestSolution(bestIndividual, results, random);
      }
    }

    private void AnalyzeBestTestSolution(IntegerVector bestIndividual, ResultCollection results, IRandom random) {
      var program = bestIndividual.ToPushProgram(config, randomPool);
      var testResult = PushEvaluator.EvaluateTest(pool, program, random);

      if (!results.ContainsKey(TestQualityResultName)) {
        results.Add(new Result(TestQualityResultName, new DoubleValue(testResult.TotalQuality)));
      } else {
        ((DoubleValue)results[TestQualityResultName].Value).Value = testResult.TotalQuality;
      }
    }

    private bool AnalyzeBestTrainingSolution(IntegerVector bestIndividual, double bestQuality, ResultCollection results, IRandom random) {
      var solution = CreatePushSolution(
        bestIndividual,
        bestQuality,
        random,
        config,
        PushEvaluator);

      if (!results.ContainsKey(BestTrainingSolutionResultName)) {
        results.Add(new Result(BestTrainingSolutionResultName, solution));
        return true;
      }

      var currentBestQuality = ((PushSolution)results[BestTrainingSolutionResultName].Value).Quality;

      if (Maximization && currentBestQuality < bestQuality ||
         !Maximization && currentBestQuality > bestQuality) {
        results[BestTrainingSolutionResultName].Value = solution;
        return true;
      }

      return false;
    }

    protected abstract PushSolution CreatePushSolution(
      IntegerVector vector,
      double bestQuality,
      IRandom random,
      IReadOnlyPushConfiguration config,
      IPushEvaluator evaluator);

    public override double Evaluate(Individual individual, IRandom random) {
      var program = individual.ToPushProgram(config, randomPool);
      var result = PushEvaluator.EvaluateTraining(pool, program, random);

      individual[CaseQualitiesScopeParameterName] = new DoubleArray(result.ExampleQualities);

      return result.TotalQuality;
    }
  }
}